(* 
app load ["Int", "Lepeget", "Listsort", "Regex", "Timer"]

val f = "test01d.txt"

*)

structure KLepeget :> KLepeget =
    (* dp01a nhf, v1.0 -- 2001. nov. 11. Hank Pter *)
struct

(* A bemenetet, ill. a kimenetet a terminlra irnythatjuk Unix alatt
   a /dev/stdin, ill. a /dev/stdout, MS-DOS alatt mindkt esetben a con
   llomnynv megadsval.
*)

exception KLepeget of string;

(* lepeget_be f = az f szveges llomnybl beolvasott adatokbl kpzett feladvny
   lepeget_be : string -> TLepeget.racspontok
*)
fun lepeget_be f =
    let

      val istr = TextIO.openIn f
      val ss = String.tokens Char.isSpace (TextIO.inputLine istr) 
                 before TextIO.closeIn istr
               (* ss = fzrknt megadott rcspontprok listja *) 

      val _ = if List.all
                 (Regex.regexecBool
                   (Regex.regcomp "\\([0-9]+,[0-9]+\\)" [Regex.Extended]) [])
                  ss
              then ()
              else
		raise KLepeget("Szintaktikusan hibs adat a(z) " ^ f ^ " llomnyban.")
              (* szintaktikusan helyes-e a feladvnyt ler llomny *)

      fun pontpar t =
            let val xys = List.mapPartial 
                            (Int.fromString o Substring.string)
                            (Regex.tokens (Regex.regcomp "[(,)]" []) t)
	    in
		(hd xys, hd(tl xys))
	    end
            (* ktelem~u listbl pontpr *)

      fun compare ((x1, y1), (x2, y2)) =
	    if      x1 > x2 then GREATER
	    else if x1 < x2 then LESS
	    else if y1 > y2 then GREATER
	    else if y1 < y2 then LESS
	    else raise KLepeget("Ismtld rcspont a(z) " ^ f ^ " llomnyban.")

    in
	Listsort.sort compare (map pontpar ss)
	  (* a feladvnyt ler llomnybl beolvasott rcspontok
	     lexikografikusan rendezett listja *) 
    end;

(* lepeget_megold (fl, mss) = az fl feladvny s az mss megoldslista alapjn
                              az sszes megoldst ellltja
   lepeget_megold : TLepeget.racspontok * TLepeget.kitoltes list -> string
*)
fun lepeget_megold (fl, mss) =
      let val _ = if length fl = length(hd mss) then () 
		  else raise KLepeget("A feladvnyt ler lista s a \
                                       \megoldslista nem egyforma hossz.")

	  fun maxPos xs = foldl Int.max 0 xs

	  val maxy = #1 (List.last fl)
	  val maxx = foldl (fn ((_,x), m) => Int.max (x,m)) 0 fl
            (* a ngyzetrcs mrete *)

	  fun negyzetracs xy fs ms =
		let val h = size(Int.toString(maxPos ms))
		    fun dash 0 = " "
		      | dash n = "-" ^ dash (n-1)
		    val hdash = dash h

		    fun nracs (0, _) _ _ = "\n"
		      | nracs (i, 0) fs ms = "\n" ^ nracs (i-1, maxx) fs ms
		      | nracs (i, j) [] _ = hdash ^ nracs (i, j-1) [] []
		      | nracs (i, j) _ [] = hdash ^ nracs (i, j-1) [] []
		      | nracs (i, j) (ffs as f::fs) (mms as m::ms) =
			  if f = (maxy+1-i, maxx+1-j)
			  then StringCvt.padLeft #" " h (Int.toString m) ^ " " ^
                                 nracs (i, j-1) fs ms
			  else hdash ^ nracs (i, j-1) ffs mms
		in
		    nracs xy fs ms
		end
      in
	  String.concat(map (negyzetracs (maxy, maxx) fl) mss)
      end

(* lepeget_ki(m, fl, mss) = az fl feladvny s az mss megoldslista alapjn
                            az sszes megoldst kirja az m szveges llomnyba.
   lepeget_ki : string * TLepeget.racspontok * TLepeget.kitoltes list -> unit
*)
fun lepeget_ki(m, fl, mss) =
      let val ostr = TextIO.openOut m
	  val _ = TextIO.output(ostr, "\n" ^ lepeget_megold(fl, mss))
      in
	 TextIO.closeOut ostr
      end

(* megold(f, m) = beolvas egy feladvnyt az f szveges llomnybl s kirja az
                  sszes megoldst az m szveges llomnyba. Ehhez termszetesen
                  szksge van a Lepeget.lepeget fggvnyre. Eredmnye az f
                  nevet, a megoldsok szmt s a futsi idt msodpercben
                  tartalmaz fzr.
   megold : string * string -> string
*)
fun megold (f, m) =
    let val fs = lepeget_be f
	val ostr = TextIO.openOut m
	fun megoldas_ki mss = 
              TextIO.output(ostr, "\n" ^ lepeget_megold(fs, mss))
	val start = Timer.startCPUTimer()
	val mss = Lepeget.lepeget fs
	val {usr=tim,...} = Timer.checkCPUTimer start
	val _ = megoldas_ki mss
	val _ = TextIO.closeOut ostr
	val t = "Table: " ^ f ^ ", " ^
                Int.toString(length mss) ^ " solution(s), " ^
		"time: " ^ Time.fmt 2 tim ^ " sec\n"
    in
	t
    end

end
